# Detection of Potential Geothermal Exploration Sites from Hyperspectral Images via Deep Learning
# Author: Ebubekir Demir <edemir@mines.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.


import csv
import math
from posixpath import split
from turtle import shape
import pandas as pd
from sklearn.cluster import KMeans
from tracemalloc import Statistic
import crop
import kmeans
from math import isnan
from statistics import mean, stdev, pstdev
from itertools import filterfalse
from pathlib import Path
import os
import pathlib
from tqdm import tqdm
import numpy as np
from osgeo import gdal, ogr
from osgeo.gdalconst import *

try:
    from osgeo import gdal_array
except:
    import gdal


KELVIN_TO_CELCIUS = -273.15
#
# Lansat 8 File extensions
#
LS8_QA_PIXEL_FE = "_QA_PIXEL.TIF"  # Quality Assessment (QA_PIXEL) band
LS8_ST_QA_FE = "_ST_QA.TIF"  # Landsat 8-9 Surface Temperature Uncertainity
LS8_ST_B10_FE = "_ST_B10.TIF"  # Surface Temeprature Band
#
# Lansat 8 File extensions Multiplicative Scale Factors
#
LS8_ST_QA_MF = 0.01  # Landsat 8-9 Surface Temperature Uncertainity
LS8_ST_B10_MF = 0.00341802  # Surface Temeprature Band
#
# Lansat 8 File extensions Additive Offsets
#
LS8_ST_B10_AO = 149.0  # Surface Temeprature Band
#
# Lansat 8 File extensions Fill Values
#
LS8_QA_PIXEL_FV = 1  # Quality Assessment (QA_PIXEL) band
LS8_ST_QA_FV = -9999  # Landsat 8-9 Surface Temperature Uncertainity
LS8_ST_B10_FV = 0  # Surface Temeprature Band


def getFileFD(directory, text):
    """
    Returns Raster file data in the directory including the text

    Parameters
    ----------
    directory : str
        The Directory where Raster file is located
    text : str
        Text that filename includes

    Returns
    -------
     Masked Array
        Numpy Masked Array extracted from raster file
    String
        Full path of the file
    """
    for filename in os.listdir(directory):
        f = os.path.join(directory, filename)
        if os.path.isfile(f):
            if text in f:
                print(f)
                try:
                    ds = gdal.Open(f)
                    band = ds.GetRasterBand(1)
                    file = band.ReadAsArray()
                    if file.dtype == np.dtype('complex128'):
                        file = file.astype('float64')
                    file = np.ma.masked_invalid(file)
                    return file, f
                except:
                    print("Error Reading File", f)
                    return None, None
    return None, None


def getFile(fullpath):
    """
    Returns Raster file data whose full path is given

    Parameters
    ----------
    fullpath : str
        Full path of the file

    Returns
    -------
     Masked Array
        Numpy Masked Array extracted from raster file
    """
    if os.path.isfile(fullpath):
        print(fullpath)
        try:
            ds = gdal.Open(fullpath)
            band = ds.GetRasterBand(1)
            file = band.ReadAsArray()
            if file.dtype == np.dtype('complex128'):
                file = file.astype('float64')
            file = np.ma.masked_invalid(file)
            return file
        except:
            print("Error Reading File", fullpath)
            return None
    return None


def saveResult(source, path, filename, res):
    """
    Returns Raster file data whose full path is given

    Parameters
    ----------
    source : str
        Full path of the source raster file
    path : str
        Directory of the raster file to be created
    filename : str
        Name of the raster file to be created
    res : Masked Array
        data of the raster file to be created

    Returns
    -------
     creates a raster file whose properties are taken from
     source raster file and data is taken from res
    """
    if not os.path.isdir(path):
        p = Path(path)
        p.mkdir(parents=True)
    if os.path.isdir(path):
        destination = os.path.join(path, filename)
        da = res.filled(fill_value=np.nan)
        mx = res.max()
        mn = res.min()
        max = str(np.round(mx, 2))
        min = str(np.round(mn, 2))
        shp = str(res.shape)
        size = str(res.size)
        invalid = str(np.ma.count_masked(res))
        percent = str(round(100.00 * np.ma.count_masked(res) / res.size, 2))
        print("WRITING    FILE: max: ", max, " min: ", min, " Shape: ", shp, " Total: ", size, " invalid: ", invalid,
              " Percent: ", percent)
        inDs = gdal.Open(source)
        band1 = inDs.GetRasterBand(1)
        rows = inDs.RasterYSize
        cols = inDs.RasterXSize
        cropData = band1.ReadAsArray(0, 0, cols, rows)
        driver = inDs.GetDriver()
        outDs = driver.Create(destination, cols, rows, 1, GDT_CFloat64)
        outBand = outDs.GetRasterBand(1)
        outBand.WriteArray(da, 0, 0)
        outBand.FlushCache()
        # outBand.SetNoDataValue(0)
        outDs.SetGeoTransform(inDs.GetGeoTransform())
        outDs.SetProjection(inDs.GetProjection())
        outDs.FlushCache()
        outDs = None
        print("W: ", destination)


def getLST(directory):
    """
    Returns Land Surface Temperature data in the folder (in Celcius)

    Parameters
    ----------
    directory : str
        The Directory where Raster file is located

    Returns
    -------
     Masked Array
        Numpy Masked Array extracted from raster file
    String
        Full path of the file
    """
    lst, lstfp = getFileFD(directory, LS8_ST_B10_FE)
    if lst is not None:
        filled = lst == LS8_ST_B10_FV
        result = (lst * LS8_ST_B10_MF) + LS8_ST_B10_AO
        # convert to celcius
        result = result + KELVIN_TO_CELCIUS
        mask = np.ma.mask_or(np.ma.getmask(result), filled)
        res = np.ma.masked_array(data=result, mask=mask)
        return res, lstfp
    return None, None


def getQAPIXEL(directory):
    """
    Returns Quality Assurance data in the folder

    Parameters
    ----------
    directory : str
        The Directory where Raster file is located

    Returns
    -------
     Masked Array
        Numpy Masked Array extracted from raster file
    String
        Full path of the file
    """
    qa, qafp = getFileFD(directory, LS8_QA_PIXEL_FE)
    if qa is not None:
        filled = qa == LS8_QA_PIXEL_FV
        mask = np.ma.mask_or(np.ma.getmask(qa), filled)
        res = np.ma.masked_array(data=qa, mask=mask)
        return res, qafp
    return None, None



def getUncertainty(directory):
    """
    Returns Uncertainty data in the folder

    Parameters
    ----------
    directory : str
        The Directory where Raster file is located

    Returns
    -------
     Masked Array
        Numpy Masked Array extracted from raster file
    String
        Full path of the file
    """
    un, unfp = getFileFD(directory, LS8_ST_QA_FE)
    if un is not None:
        filled = un == LS8_ST_QA_FV
        mask = np.ma.mask_or(np.ma.getmask(un), filled)
        res = np.ma.masked_array(data=un, mask=mask)
        return res, unfp
    return None, None


def qa_analysis(source, destination, destext):
    b = os.path.isdir(source)
    print(b)
    for x in os.walk(source):
        splt = x[0].split("\\")
        lst, fp = getLST(x[0])
        qa, qafp = getQAPIXEL(x[0])

        if lst is not None and qa is not None:
            # Clear with lows set
            filled = qa == 21824
            mask = np.ma.mask_or(np.ma.getmask(lst), filled)
            res = np.ma.masked_array(data=lst, mask=mask)
            print("QA ANALYSIS: LST max: ", str(round(lst.max(), 2)), " min: ", str(round(lst.min(), 2)), " Shape: ",
                  str(lst.shape), " Total: ", str(lst.size), " invalid: ", str(np.ma.count_masked(lst)), " Percent: ",
                  str(round(100.00 * np.ma.count_masked(lst) / lst.size, 2)))
            print("QA ANALYSIS: RES max: ", str(round(res.max(), 2)), " min: ", str(round(res.min(), 2)), " Shape: ",
                  str(res.shape), " Total: ", str(res.size), " invalid: ", str(np.ma.count_masked(res)), " Percent: ",
                  str(round(100.00 * np.ma.count_masked(res) / res.size, 2)))
            dest = x[0].replace(source, destination)
            filename = splt[-1] + destext
            saveResult(fp, dest, filename, res)



def export_analysis_qa(filepath, destination, destext):
    """
    Returns statistical analysis of files in destination folder and subfolders

    Parameters
    ----------
    filepath : str
        Full path of the csv file to be created
    destination : str
        folder that files to be analysed
    destext : str
        extension of files to be analysed

    Returns
    -------
     list
        list of all files with statistical results
    """
    list = []
    ls = []
    ls.append("filename")
    ls.append("X")
    ls.append("Y")
    ls.append("Total")
    ls.append("Valid")
    ls.append("Percent")
    ls.append("max")
    ls.append("min")
    ls.append("range")
    ls.append("mean")
    ls.append("stdev")
    list.append(ls)
    for p in destination:
        for x in os.walk(p):
            for y in x[2]:
                if destext in y:
                    f = os.path.join(x[0], y)
                    file = getFile(f)
                    valid = np.ma.count(file)
                    if file is not None and valid != 0:
                        ls = []
                        total = file.size
                        percent = round(100 * valid / total, 2)
                        max = round(file.max(), 2)
                        min = round(file.min(), 2)
                        range = round(max - min, 2)
                        mean = round(file.mean(), 2)
                        stdev = round(file.std(), 2)
                        print(y, ",", file.shape, ",", total, ",", valid, ",", percent, ",", max, ",", min, ",", range,
                              ",", mean, ",", stdev)
                        ls.append(y)
                        ls.append(file.shape[0])
                        ls.append(file.shape[1])
                        ls.append(total)
                        ls.append(valid)
                        ls.append(percent)
                        ls.append(max)
                        ls.append(min)
                        ls.append(range)
                        ls.append(mean)
                        ls.append(stdev)
                        list.append(ls)
                    elif file is None:
                        print("File is None:", y)
                    elif valid == 0:
                        print("No valid pixels in file", y)
                    else:
                        print("UNKNOWN ERROR")
    df = pd.DataFrame(list)
    df.to_csv(filepath, header=False, index=False)
    return list


def crop_uncertainity_files(cropfile, source, sourcetext, replacetext, destext):
    """
    Crops Uncertainty Files and saves it to

    Parameters
    ----------
    filepath : str
        Full path of the csv file to be created
    destination : str
        folder that files to be analysed
    destext : str
        extension of files to be analysed

    Returns
    -------
     list
        list of all files with statistical results

    Example Usage
    """
    for p in source:
        for x in os.walk(p):
            unc, uncfd = getUncertainty(x[0])
            if unc is not None:
                dest = uncfd.replace(sourcetext, replacetext)
                dest = dest.replace("DOT_TIF", destext)
                crop.main(cropfile, uncfd, dest)


def crop_files(cropfile, source, sourceext, destext):
    for p in source:
        for x in os.walk(p):
            file, filefd = getFileFD(x[0], sourceext)
            if file is not None:
                dest = filefd.replace(sourceext, destext)
                crop.main(cropfile, filefd, dest)


def remove_uncertain_pixels(path, uc_level, uncext, lstext, destext):
    for p in path:
        for x in os.walk(p):
            lst, lstfd = getFileFD(x[0], lstext)
            unc, uncfd = getFileFD(x[0], uncext)
            if lst is not None and unc is not None:
                filled = unc >= (uc_level / LS8_ST_QA_MF)
                mask = np.ma.mask_or(np.ma.getmask(lst), filled)
                res = np.ma.masked_array(data=lst, mask=mask)
                validcount = np.ma.count(res)
                if validcount != 0:
                    dext = destext + str(uc_level) + "DOT_TIF"
                    dest = lstfd.replace(lstext, dext)
                    split = dest.split("\\")
                    saveResult(lstfd, x[0], split[-1], res)
                else:
                    print("NOT WRITE No valid pixels:", lstfd, uncfd)


def statistics(path, destext):
    list = []
    for x in os.walk(path):
        for y in x[2]:
            if destext in y:
                f = os.path.join(x[0], y)
                file = getFile(f)
                valid = np.ma.count(file)
                if file is not None and valid != 0:
                    ls = []
                    total = file.size
                    percent = 100 * valid / total
                    max = file.max()
                    min = file.min()
                    range = max - min
                    mean = file.mean()
                    stdev = file.std()
                    ls.append(file)
                    ls.append(y)
                    ls.append(valid)
                    ls.append(mean)
                    ls.append(stdev)
                    list.append(ls)
    return list


def kmean(destfile, path, ext):
    fl = open(destfile, "w")
    writer = csv.writer(fl)
    for p in path:
        s = statistics(p, ext)
        for sl in tqdm(s):
            l1 = sl[0]
            mn = sl[3]
            st = sl[4]
            fn = sl[1]
            print(sl[1])
            if st != 0:
                f = l1.flatten()
                filtered = [x for x in f if x != np.nan]
                filtered = (filtered - mn) / st
                if filtered.size > 30:
                    res, lst = kmeans.elb_run(np.array(filtered))
                    writer.writerow((fn, res, lst))
                else:
                    print("Not enough SIZE to start kmean:", filtered.size)
            else:
                print("Not enough STANDARD DEVIATION to start kmean:", st)
    fl.close()


def getk(data, value):
    for index, row in data.iterrows():
        name = row[0]
        if not row.hasnans:
            name = row[0]
            if value in name:
                return row[1]
    return None


def kmeanscalculateandsave(csvsource, destination, extension, name, kmext):
    csvfile = open(csvsource, newline='')
    csvdata = csv.reader(csvfile, delimiter=',', quotechar='|')
    d = pd.DataFrame(csvdata)
    for p in destination:
        for x in os.walk(p):
            splt = x[0].split("\\")
            coso, filename = getFileFD(x[0], extension)
            k = getk(d, splt[-1])
            if coso is not None and k is not None:
                dt = coso.data
                ls = []
                for i in range(0, dt.shape[0]):
                    for j in range(0, dt.shape[1]):
                        val = dt[i][j]
                        if not math.isnan(val):
                            ls.append((i, j, val))
                dtarray = np.array(ls)
                m = dtarray.max()
                n = dtarray.min()
                data = dtarray[:, 2]
                data = data.reshape(-1, 1)
                k = int(k)
                km = KMeans(n_clusters=k, max_iter=500).fit(data)
                ind = np.argsort(km.cluster_centers_, axis=None)
                param = ind / (k - 1)
                fd = dt.copy()
                fd[:] = np.nan
                for kk in range(0, km.labels_.size):
                    ii = dtarray[kk][0]
                    jj = dtarray[kk][1]
                    for s in range(0, k):
                        if km.labels_[kk] == ind[s]:
                            fd[int(ii)][int(jj)] = param[s]
                            break
                fn = splt[-1] + name + str(k) + kmext
                fd = np.ma.masked_invalid(fd)
                saveResult(filename, x[0], fn, fd)


def kmeanscombine(path, kmext, dest, combined, norm, threshold):
    list = []
    source = None
    data = None
    for p in path:
        for x in os.walk(p):
            splt = x[0].split("\\")
            kmfile, filename = getFileFD(x[0], kmext)
            if kmfile is not None:
                valid = kmfile.count()
                total = kmfile.data.size
                percent = 100.00 * valid / total
                if percent > threshold:
                    list.append(kmfile)
                    source = filename
                    data = kmfile

    ls = np.array(list)
    res = np.nanmean(ls, axis=0)
    fd = np.ma.masked_invalid(res)
    saveResult(source, dest, combined, fd)
    fd = (fd - fd.min()) / (fd.max() - fd.min())
    saveResult(source, dest, norm, fd)


if __name__ == '__main__':
    # example usages
    crop_uncertainity_files("", [], "", "", "")
    # crop_files("", [], "", "", "")
    # export_analysis_qa("", [], "")
    # for i in range(2,12):
    #     remove_uncertain_pixels([], i, "", "", "")
    # for i in range(2,12):
    #     kmean("", [], "")
    # for i in range(2,12):
    #     kmeanscalculateandsave("", [], "", "")
    # for i in range(2,12):
    #     export_analysis_qa("", [], "") 
    # for i in range(2, 12):
    #     kmeanscombine([], "", "", "", 0)
