# Detection of Potential Geothermal Exploration Sites from Hyperspectral Images via Deep Learning
# Author: Ebubekir Demir <edemir@mines.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.


from __future__ import print_function
import time

start = time.time()
from datetime import timedelta
from unicodedata import name
import numpy as np
import doe_tiff2 as dt
import os
import argparse
from tqdm import tqdm

# import the necessary packages
import keras
import tensorflow as tf
from keras.callbacks import EarlyStopping, ModelCheckpoint
from keras.layers import AveragePooling2D, Input, Concatenate
from keras.layers import Conv2D, MaxPooling2D
from keras.layers import Activation, Flatten, Dropout, Dense
from keras.models import Model
from tensorflow.keras.optimizers import Adam, Nadam, Adadelta, Adagrad, Adamax, SGD
from keras.regularizers import l2
from tensorflow.keras.utils import to_categorical

import matplotlib
import matplotlib.pyplot as plt
from numpy import argmax
import os
import os.path as path
import sys
import pickle
import random
import skimage
from sklearn.preprocessing import LabelBinarizer
from sklearn.model_selection import train_test_split
from sklearn.metrics import classification_report, confusion_matrix, accuracy_score
from sklearn.metrics import roc_curve, auc

# Default values
SAMPLES_TO_CREATE = 1200
KERNEL_PIXELS = 17
KERNEL_CHANNELS = 5
DEFAULT_DIRECTORY = 'dataset'

matplotlib.use("Agg")
DEFAULT_WEIGHTS_FILE_NAME = 'doe_cnn.weights'
DEFAULT_MODEL_FILE_NAME = 'doe_cnn.model'
DEFAULT_LABEL_FILE_NAME = 'doe_cnn.labels.pickle'

INIT_LR = 5e-1  # Default Loss Rate
INIT_DECAY = 1e-3  # Default Decay
CHANNELS = 5  # This will be redefined based on parameters
KERNEL_PIXELS = 17  # Default pixels by side on each tile
NUM_CLASSES = 2  # Default number of classes ("Geothermal", "Non-Geothermal")
BS = 32  # Default batch size
EPOCHS = 500

class2code = {'none': 0,
              'Non-geothemal': 1,
              'Geothemal': 2}
code2class = {0: 'none',
              1: 'Non-geothemal',
              2: 'Geothemal'}

try:
    raw_input  # Python 2
except NameError:
    raw_input = input  # Python 3

''' Class definitions '''


class doe_ann_object(object):
    def __init__(self, root_dir):
        self.root_dir = root_dir


class DataGenerator(keras.utils.Sequence):
    'Generates data for Keras'

    def __init__(self, data_set, labels, batch_size=BS,
                 dim=(KERNEL_PIXELS, KERNEL_PIXELS, CHANNELS),
                 n_channels=CHANNELS, n_classes=NUM_CLASSES,
                 shuffle=True, augment_data=False):
        'Initialization'
        self.dim = dim
        self.batch_size = batch_size
        self.labels = labels
        self.data_set = data_set
        self.n_channels = n_channels
        self.n_classes = n_classes
        self.shuffle = shuffle
        self.augment_data = augment_data
        self.on_epoch_end()

    def __len__(self):
        'Denotes the number of batches per epoch'
        return int(np.floor(len(self.data_set) / self.batch_size))

    def __getitem__(self, index):
        'Generate one batch of data'
        # Generate indexes of the batch
        indexes = self.indexes[index * self.batch_size:(index + 1) * self.batch_size]
        # Generate data
        X, y = self.__data_generation(indexes)

        return X, y

    def on_epoch_end(self):
        'Updates indexes after each epoch'
        self.indexes = np.arange(len(self.data_set))
        if self.shuffle == True:
            np.random.shuffle(self.indexes)

    def __data_generation(self, list_IDs_temp):
        'Generates data containing batch_size samples'  # X : (n_samples, *dim, n_channels)
        # Initialization
        X = np.empty((self.batch_size, *self.dim, self.n_channels))
        y = np.empty((self.batch_size, self.n_classes), dtype=int)
        # Generate data
        for i, ID in enumerate(list_IDs_temp):
            # Store sample
            image_raw = self.data_set[ID]
            if self.augment_data:
                if np.random.randint(1):
                    image_raw = np.rot90(image_raw, 2)
                if np.random.randint(1):
                    image_raw = np.fliplr(image_raw)
                if np.random.randint(1):
                    image_raw = np.flipud(image_raw)
            X[i,] = image_raw
            # Store class
            y[i] = self.labels[ID]
        return X, y


def getj(val, cnv1):
    jr = Conv2D(16, (1, 1), padding='same', activation='relu', kernel_regularizer=l2(0.002))(cnv1)
    j = Conv2D(64, (val, val), padding='same', activation='relu', kernel_regularizer=l2(0.002))(jr)
    return j


def jigsaw_m(inp, size, first_layer=None):
    print("JIGSAW_M:", inp, size)
    cnv1 = Conv2D(128, (1, 1), padding='same', activation='relu', kernel_regularizer=l2(0.002))(inp)
    j1x1 = Conv2D(256, (1, 1), padding='same', activation='relu', kernel_regularizer=l2(0.002))(cnv1)

    j3x3r = Conv2D(96, (1, 1), padding='same', activation='relu', kernel_regularizer=l2(0.002))(cnv1)
    j3x3 = Conv2D(128, (3, 3), padding='same', activation='relu', kernel_regularizer=l2(0.002))(j3x3r)

    j5x5r = Conv2D(16, (1, 1), padding='same', activation='relu', kernel_regularizer=l2(0.002))(cnv1)
    j5x5 = Conv2D(128, (5, 5), padding='same', activation='relu', kernel_regularizer=l2(0.002))(j5x5r)

    j7x7r = Conv2D(16, (1, 1), padding='same', activation='relu', kernel_regularizer=l2(0.002))(cnv1)
    j7x7 = Conv2D(128, (7, 7), padding='same', activation='relu', kernel_regularizer=l2(0.002))(j7x7r)

    j9x9r = Conv2D(16, (1, 1), padding='same', activation='relu', kernel_regularizer=l2(0.002))(cnv1)
    j9x9 = Conv2D(64, (9, 9), padding='same', activation='relu', kernel_regularizer=l2(0.002))(j9x9r)

    ls = [j1x1, j3x3, j5x5, j7x7, j9x9]
    print("JIGSAW_M:", ls)
    for ks in range(11, size + 1, 2):
        print("JIGSAW_M:", ks)
        ls.append(getj(ks, cnv1))
    print("JIGSAW_M:", ls)
    jigsaw_t1_pool = MaxPooling2D(pool_size=(3, 3), strides=(1, 1), padding='same', name="jigsaw_t1_pool")(cnv1)
    jigsaw_t1_pool_proj = Conv2D(32, (1, 1), padding='same', activation='relu', kernel_regularizer=l2(0.002))(
        jigsaw_t1_pool)
    if first_layer is None:
        ls.append(jigsaw_t1_pool_proj)
        print("JIGSAW_M:", ls)
        jigsaw_t1_output = Concatenate(axis=-1, name="jigsaw_t1_output")(ls)
    else:
        jigsaw_t1_first = Conv2D(96, (1, 1), padding='same', activation='relu', kernel_regularizer=l2(0.002),
                                 name="jigsaw_t1_output")(first_layer)
        ls.append(jigsaw_t1_first).append(jigsaw_t1_pool_proj)
        print("JIGSAW_M:", ls)
        jigsaw_t1_output = Concatenate(axis=-1, name="jigsaw_t1_output")(ls)
    return jigsaw_t1_output


def jigsaw_m_end(input_net, num_classes=NUM_CLASSES, first_layer=None):
    avg_pooling = AveragePooling2D(pool_size=(3, 3), strides=(1, 1), name='avg_pooling')(input_net)
    flat = Flatten(name="flat")(avg_pooling)
    flat = Dense(16, kernel_regularizer=l2(0.002), name="dense16")(flat)
    flat = Dropout(0.4, name="droput")(flat)
    if first_layer is not None:
        input_pixel = Flatten()(first_layer)
        input_pixel = Dense(16, kernel_regularizer=l2(0.002))(input_pixel)
        input_pixel = Dropout(0.2)(input_pixel)
        input_pixel = Dense(16, kernel_regularizer=l2(0.002))(input_pixel)
        input_pixel = Dropout(0.2)(input_pixel)
        flat = Concatenate(axis=-1, name="concat1")([input_pixel, flat])
    flat = Dense(32, kernel_regularizer=l2(0.002), name="dense32")(flat)
    avg_pooling = Dropout(0.4)(flat)
    print('num classes: ', num_classes)
    loss3_classifier = Dense(num_classes, kernel_regularizer=l2(0.002), name="denselc")(avg_pooling)
    loss3_classifier_act = Activation('softmax', name='prob')(loss3_classifier)
    return loss3_classifier_act


def ROC_curve_calc(testY, pre_y2, class_num, output_file_header):
    '''
    Calculate other stats
    '''
    fpr = dict()
    tpr = dict()
    roc_auc = dict()
    n_classes = class_num
    for i in range(1, n_classes):
        class_plot = i
        true_labels = (testY[:] == i)
        pred_probs = (pre_y2[:] == i)
        if class_plot == 0:
            pred_probs = pred_probs * 0
            pred_probs = pre_y2_prob[:, 0] * 0
        else:
            pred_probs = pre_y2_prob[:, class_plot - 1]
        print('For class ', i, ': shape = ', true_labels.shape)
        print('For true_labels, first item = ', true_labels[0])
        fpr[i], tpr[i], _ = roc_curve(true_labels, pred_probs)  # , pos_label = 1)
        roc_auc[i] = auc(fpr[i], tpr[i])
    for i in range(1, n_classes):
        # Plot for a class
        class_plot = i
        class_name = code2class[class_plot]
        plt.figure()
        line_width = 2
        plt.plot(fpr[class_plot], tpr[class_plot], color='darkorange',
                 lw=line_width, label='ROC Curve for class %s (area = %0.2f)' % (class_name, roc_auc[class_plot]))
        plt.plot([0, 1], [0, 1], color='navy', lw=line_width, linestyle='--')
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title('ROC Curve')
        plt.legend(loc='lower right')
        plt.show()
        plt.savefig(output_file_header + '.' + class_name + '.png')
    return


print('Set-up complete.')

if __name__ == '__main__':

    ''' Main instructions '''
    print('Parsing input...')
    ap = argparse.ArgumentParser()

    ap.add_argument("-i", "--image", required=True, help="path to input multi-band image (i.e., image file name)")
    ap.add_argument("-i2", "--image2", required=False,
                    help="path to second input multi-band image (i.e., image file name) (optional)", default=None)
    ap.add_argument("-i3", "--image3", required=False,
                    help="path to third input multi-band image (i.e., image file name) (optional)", default=None)

    ap.add_argument("-s", "--samples", required=True, help="number of samples to create", type=int)
    ap.add_argument("-b", "--batch_size", required=True, help="Defines batch size", type=int)
    ap.add_argument("-e", "--epochs", required=True, type=int, help="Number of epochs to train")

    ap.add_argument("-c", "--channels", required=False, help='Number of channels in each image', default=CHANNELS,
                    type=int)
    ap.add_argument("-g", "--gpus", required=False, type=int, help="Number of GPUs to run in parallel", default=1)

    ap.add_argument("-k", "--kernel_size", required=False, help='Number of pixels by side in each image',
                    default=KERNEL_PIXELS, type=int)
    ap.add_argument("-l", "--labelbin", required=True, help="path to output label binarizer(l)")
    ap.add_argument("-m", "--model", required=True, help="path to output model (h5)")
    ap.add_argument("-p", "--plot", required=True, help="Plot file (png)")

    args = vars(ap.parse_args())

    image_name = args["image"]
    image_name2 = args["image2"]
    image_name3 = args["image3"]

    num_samples = args["samples"]
    batch_size = args["batch_size"]
    num_epochs = args["epochs"]

    num_channels = args["channels"]
    image_channels = num_channels - 1
    num_gpus = args["gpus"]

    kernel_size = args["kernel_size"]
    kernel_pixels = kernel_size
    label_file = args["labelbin"]
    model_file = args["model"]
    plot_file = args["plot"]

    augment_data = None
    reset_model = "reset"
    num_classes = NUM_CLASSES
    true_random = None
    validate_only = None
    weights_file = None
    output_curves_file = None

    PADDING = int(kernel_size / 2)
    # img_b = dt.io.read_gdal_file(image_name)
    img_b = np.load(image_name)
    max_channels = img_b.shape[2]
    assert num_channels > 0, 'Channels has to be a positive integer'
    assert num_channels <= max_channels, 'Channels (' + str(num_channels) + ') has to be equal or lower than ' + str(
        max_channels)
    img_b_scaled = img_b[:, :, 1:num_channels + 1]
    print(img_b_scaled.shape)
    mask_b = img_b[:, :, 0]  # first band is Ground Truth
    for i in range(0, img_b_scaled.shape[2]):
        print("band (", i, ") min:", img_b_scaled[:, :, i].min())
        print("band (", i, ") max:", img_b_scaled[:, :, i].max())
    img_b_scaled = dt.frame_image(img_b_scaled, PADDING)
    sc = dt.GeoTiffConvolution(img_b_scaled, kernel_size, kernel_size)

    new_map = np.zeros_like(mask_b)  # creates empty map
    data = []
    labels = []
    # Land classes go from 0 to 1
    for land_type_class in range(0, 2):
        mask_b_class = np.array(np.where(mask_b == land_type_class)).T
        print('Items in class ', land_type_class, ': ', len(mask_b_class))
        mask_b_scaled = mask_b_class + [PADDING, PADDING]
        ln = len(mask_b_scaled)
        choices = np.random.choice(len(mask_b_scaled), num_samples, replace=False)
        for c in tqdm(choices, desc="Calculating...", ascii=False, ncols=75):
            x = mask_b_scaled[c][0] - PADDING
            y = mask_b_scaled[c][1] - PADDING
            new_map[x, y] = 1 + land_type_class
            my_slice = sc.apply_mask(mask_b_scaled[c][0], mask_b_scaled[c][1])
            data.append(my_slice)
            labels.append(land_type_class)

    if image_name2 is not None:
        # img_b = dt.io.read_gdal_file(image_name2)
        img_b = np.load(image_name2)
        max_channels = img_b.shape[2]
        assert num_channels > 0, 'Channels has to be a positive integer'
        assert num_channels <= max_channels, 'Channels has to be equal or lower than {}'.format(max_channels)
        img_b_scaled = img_b[:, :, 1:num_channels + 1]
        mask_b = img_b[:, :, 0]  # first band is Ground Truth
        for i in range(0, img_b_scaled.shape[2]):
            print("band (", i, ") min:", img_b_scaled[:, :, i].min())
            print("band (", i, ") max:", img_b_scaled[:, :, i].max())
        img_b_scaled = dt.frame_image(img_b_scaled, PADDING)
        sc = dt.GeoTiffConvolution(img_b_scaled, kernel_size, kernel_size)

        new_map = np.zeros_like(mask_b)  # creates empty map
        # Land classes go from 0 to 1
        for land_type_class in range(0, 2):
            mask_b_class = np.array(np.where(mask_b == land_type_class)).T
            print('Items in class ', land_type_class, ': ', len(mask_b_class))
            mask_b_scaled = mask_b_class + [PADDING, PADDING]
            ln = len(mask_b_scaled)
            choices = np.random.choice(len(mask_b_scaled), num_samples, replace=False)
            for c in tqdm(choices, desc="Calculating...", ascii=False, ncols=75):
                x = mask_b_scaled[c][0] - PADDING
                y = mask_b_scaled[c][1] - PADDING
                new_map[x, y] = 1 + land_type_class
                my_slice = sc.apply_mask(mask_b_scaled[c][0], mask_b_scaled[c][1])
                data.append(my_slice)
                labels.append(land_type_class)

    if image_name3 is not None:
        # img_b = dt.io.read_gdal_file(image_name3)
        img_b = np.load(image_name3)
        max_channels = img_b.shape[2]
        assert num_channels > 0, 'Channels has to be a positive integer'
        assert num_channels <= max_channels, 'Channels has to be equal or lower than {}'.format(max_channels)
        img_b_scaled = img_b[:, :, 1:num_channels + 1]
        mask_b = img_b[:, :, 0]  # first band is Ground Truth
        for i in range(0, img_b_scaled.shape[2]):
            print("band (", i, ") min:", img_b_scaled[:, :, i].min())
            print("band (", i, ") max:", img_b_scaled[:, :, i].max())
        img_b_scaled = dt.frame_image(img_b_scaled, PADDING)
        sc = dt.GeoTiffConvolution(img_b_scaled, kernel_size, kernel_size)

        new_map = np.zeros_like(mask_b)  # creates empty map
        # Land classes go from 0 to 1
        for land_type_class in range(0, 2):
            mask_b_class = np.array(np.where(mask_b == land_type_class)).T
            print('Items in class ', land_type_class, ': ', len(mask_b_class))
            mask_b_scaled = mask_b_class + [PADDING, PADDING]
            ln = len(mask_b_scaled)
            choices = np.random.choice(len(mask_b_scaled), num_samples, replace=False)
            for c in tqdm(choices, desc="Calculating...", ascii=False, ncols=75):
                x = mask_b_scaled[c][0] - PADDING
                y = mask_b_scaled[c][1] - PADDING
                new_map[x, y] = 1 + land_type_class
                my_slice = sc.apply_mask(mask_b_scaled[c][0], mask_b_scaled[c][1])
                data.append(my_slice)
                labels.append(land_type_class)

    IMAGE_DIMS = (kernel_pixels, kernel_pixels, image_channels)
    BATCH_DIMS = (None, kernel_pixels, kernel_pixels, image_channels)

    print('Read images:', len(data))
    # scale the raw pixel intensities to the range [0, 1]
    data = np.asarray(data, dtype=np.float32)
    data = np.nan_to_num(data)
    labels = np.array(labels)
    print("[INFO] data matrix: {:.2f}MB".format(
        data.nbytes / (1024 * 1024.0)))
    # binarize the labels
    lb = LabelBinarizer()
    labels_lb = lb.fit_transform(labels)

    y_binary = to_categorical(labels, num_classes=num_classes)
    y_inverse = argmax(y_binary, axis=1)
    print('Read labels:', (lb.classes_))
    print('Transformed labels:', (y_binary[:10]))
    print('Inverted labels:', (y_inverse[:10]))
    print('Any nulls?: ', np.isnan(data).any())
    ###
    '''
    Creates the network
    '''
    # Define distribution strategy
    strategy = tf.distribute.MirroredStrategy()

    # construct model under distribution strategy scope
    with strategy.scope():

        my_input = Input(shape=IMAGE_DIMS, name="input")
        # One jigsaw module(s)
        jigsaw_01 = jigsaw_m(my_input, kernel_size)
        # Attaches end to jigsaw modules, returns class within num_classes
        loss3_classifier_act = jigsaw_m_end(jigsaw_01, num_classes=num_classes,
                                            first_layer=my_input)  # testing num_classes
        # Builds model
        model3 = Model(inputs=my_input, outputs=[loss3_classifier_act])
        model3.summary()
        # partition the data into training and testing splits using 20% of
        # the data for training and the remaining 80% for testing/validation
        (trainX, testX, trainY, testY) = train_test_split(data,
                                                          y_binary, test_size=0.8, random_state=42)
        (validateX, testX, validateY, testY) = train_test_split(testX, testY,
                                                                test_size=0.75, random_state=42)
        params = {'dim': (kernel_pixels, kernel_pixels), 'batch_size': batch_size,
                  'n_classes': num_classes,
                  'n_channels': image_channels,
                  'shuffle': True, 'augment_data': augment_data}
        # construct the image generator for data augmentation
        my_batch_gen = DataGenerator(trainX, trainY, **params)
        print('creating generator with trainX, trainY of shapes: (%s, %s)' %
              (trainX.shape, trainY.shape)
              )
        ## initialize the model
        print("[INFO] compiling model...")
        print('SmallJigsaw: (depth, width, height, classes) = (%s, %s, %s, %s)' %
              (IMAGE_DIMS[0], IMAGE_DIMS[1], IMAGE_DIMS[2], len(lb.classes_))
              )
        if (validate_only):
            print("[INFO] Skipping training...")
            print("[INFO] Validate-only model:", validate_only)
            pass
        else:
            print("[INFO] Training...")
            print("[INFO] Reset model:", reset_model)
            print("[INFO] Validate-only model:", validate_only)
            # Hyperparameter selection: Optimization function
            # opt = Adam(lr=INIT_LR, decay=INIT_DECAY)   # Old decay was: INIT_LR / EPOCHS)
            # opt = Adam(lr=INIT_LR, beta_1=INIT_DECAY, amsgrad=True)
            # opt = Adadelta(learning_rate=INIT_LR)
            opt = Adadelta()
            model3.compile(loss="categorical_crossentropy", optimizer=opt, metrics=["accuracy"])
            # define the network's early stopping
            print("[INFO] define early stop and auto save for network...")
            auto_save = ModelCheckpoint(model_file, monitor='val_accuracy', verbose=0,
                                        save_best_only=True, save_weights_only=False,
                                        mode='auto')
            # can use validation set loss or accuracy to stop early
            # early_stop = EarlyStopping( monitor = 'val_accuracy', mode='max', baseline=0.97)
            # patience was 50
            early_stop = EarlyStopping(monitor='val_loss', mode='min', verbose=1, patience=10)
            # train the network
            print("[INFO] training network...")
            # Train the model
            H = model3.fit(
                my_batch_gen,
                validation_data=(validateX, validateY),
                steps_per_epoch=len(trainX) // batch_size,
                callbacks=[auto_save, early_stop],
                epochs=num_epochs, verbose=1)
            # save the model to disk
            print("[INFO] serializing network...")
            model3.save(model_file)
            # save the label binarizer to disk
            print("[INFO] serializing label binarizer...")
            f = open(label_file, "wb")
            f.write(pickle.dumps(lb))
            f.close()
        testY = argmax(testY, axis=1)
        print('[INFO] Model Info ...')
        print("summary: ", model3.summary())
        print("get_weights: ", model3.get_weights())
        print('[INFO] Predicting ...')
        pre_y2 = model3.predict(testX, verbose=1)
        print("pred set:", pre_y2[:10])
        pre_y2_prob = pre_y2
        pre_y2 = pre_y2.argmax(axis=-1)
        acc2 = accuracy_score(testY, pre_y2)
        print("test set:", testY[:10])
        print("pred set:", pre_y2[:10])
        print('Accuracy on test set: {0:.3f}'.format(acc2))
        print("Confusion Matrix:")
        print(confusion_matrix(testY, pre_y2))
        print()
        print("Classification Report")
        print(classification_report(testY, pre_y2))
        # Calculate ROC Curves if required
        if output_curves_file is not None:
            ROC_curve_calc(testY, pre_y2, class_num=8, output_file_header=output_curves_file)
        if (not validate_only):
            # plot the training loss and accuracy
            plt.style.use("ggplot")
            plt.figure()
            N = num_epochs
            plt.plot(np.arange(0, N), H.history["loss"], label="train_loss")
            plt.plot(np.arange(0, N), H.history["val_loss"], label="val_loss")
            plt.plot(np.arange(0, N), H.history["accuracy"], label="train_acc")
            plt.plot(np.arange(0, N), H.history["val_accuracy"],
                     label="val_accuracy")
            plt.title("Training Loss and Accuracy")
            plt.xlabel("Epoch #")
            plt.ylabel("Loss/Accuracy")
            plt.legend(loc="upper left")
            plt.savefig(plot_file)

end = time.time()
td = timedelta(seconds=int(end - start))
print("Total processing time", td)
