# Detection of Potential Geothermal Exploration Sites from Hyperspectral Images via Deep Learning
# Author: Ebubekir Demir <edemir@mines.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

import argparse

from osgeo import gdal, ogr, osr
import numpy as np
import pandas as pd
from tqdm import tqdm
import math
from osgeo import gdal
from osgeo.gdalconst import *


def distancemap(groundtruth, result, value):
    ds = gdal.Open(groundtruth)
    xsize = ds.RasterXSize
    ysize = ds.RasterYSize
    gnd_layer = ds.GetRasterBand(1)
    gnd = gnd_layer.ReadAsArray()
    p_data = np.pad(gnd, (1,), 'constant', constant_values=0)
    south = p_data.copy()
    south = south[2: 2 + ysize, 1: 1 + xsize]
    north = p_data.copy()
    north = north[0: 0 + ysize, 1: 1 + xsize]
    east = p_data.copy()
    east = east[1: 1 + ysize, 2: 2 + xsize]
    west = p_data.copy()
    west = west[1: 1 + ysize, 0: 0 + xsize]

    gndd = gnd == value
    east = east != value
    west = west != value
    north = north != value
    south = south != value

    res = np.logical_or(east, west)
    res = np.logical_or(res, north)
    res = np.logical_or(res, south)
    res = np.logical_and(gndd, res)

    gndarr = np.array(np.where(res == True)).T
    ind = np.indices((ysize, xsize))
    res = np.ones((ysize, xsize)) * np.inf
    for y, x in tqdm(gndarr):
        yind = ind[0] - y
        xind = ind[1] - x
        res1 = np.empty((2, ysize, xsize))
        res1[0] = yind
        res1[1] = xind
        nm = np.linalg.norm(res1, axis=0)
        res1[0] = res
        res1[1] = nm
        res = np.min(res1, axis=0)

    res[gnd == value] = 0

    driver = ds.GetDriver()
    outDs = driver.Create(result, xsize, ysize, 1, GDT_CFloat32)
    outBand = outDs.GetRasterBand(1)
    outData = res
    outBand.WriteArray(outData, 0, 0)
    outBand.FlushCache()
    outBand.SetNoDataValue(-99)
    outDs.SetGeoTransform(ds.GetGeoTransform())
    outDs.SetProjection(ds.GetProjection())
    outDs.FlushCache()
    del outData
    outDs = None


if __name__ == '__main__':
    ap = argparse.ArgumentParser()
    ap.add_argument("-d", "--data", required=True, help="path to second input multi-band im")
    ap.add_argument("-o", "--output", required=True, help="path to third input multi-band image")
    ap.add_argument("-v", "--value", required=True, help="start")
    args = vars(ap.parse_args())
    data = args["data"]
    out = args["output"]
    s = int(args["value"])
    distancemap(data, out, s)
