# Detection of Potential Geothermal Exploration Sites from Hyperspectral Images via Deep Learning
# Author: Ebubekir Demir <edemir@mines.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.


import numpy as np
from osgeo import gdal
from tqdm import tqdm
import os


def createDensity(map, sourcefolder, destinationfolder, model):
    PRE = "path" + map + "model_prefix" + model + ".gri"
    ds = gdal.Open(PRE)
    p_data = removecomplexpart(ds.GetRasterBand(1).ReadAsArray().T)
    xai = ["lrp.z", "input_t_gradient", "guided_backprop", "gradient", "deep_taylor", "deconvnet"]
    for x in xai:
        NPY = sourcefolder + map + "-" + x + "model_prefix" + model + ".npy"
        dt = np.load(NPY)
        dt = np.array(dt, dtype=np.float32)
        data_m = dt[:, :, 0]
        data_m[p_data != 1] = np.nan
        data_t = dt[:, :, 1]
        data_t[p_data != 1] = np.nan
        data_f = dt[:, :, 2]
        data_f[p_data != 1] = np.nan
        data = np.empty(data_m.shape)
        data_mf = np.max(np.stack((data_m, data_f)), axis=0)
        data_ft = np.max(np.stack((data_t, data_f)), axis=0)
        data_mt = np.max(np.stack((data_m, data_t)), axis=0)
        data[data_m > data_ft] = 1
        data[data_t > data_mf] = 2
        data[data_f > data_mt] = 3
        data[(data_f == data_t) & (data_f > data_m)] = 4
        data[(data_f == data_m) & (data_f > data_t)] = 5
        data[(data_m == data_t) & (data_t > data_f)] = 6
        data[(data_f == data_t) & (data_f == data_m)] = 7
        data[p_data != 1] = np.nan
        c, v = np.unique(data, return_counts=True)
        xsize = ds.RasterXSize
        ysize = ds.RasterYSize
        gettr = ds.GetGeoTransform()
        crs = ds.GetProjection()

        RES = destinationfolder + map + "-" + x + "model_prefix" + model + ".gri"
        target_ds = gdal.GetDriverByName('GTiff').Create(RES, xsize, ysize, 1, gdal.GDT_Int32)
        target_ds.SetGeoTransform(gettr)
        target_ds.SetProjection(crs)
        target_band = target_ds.GetRasterBand(1)
        target_band.WriteArray(data.T, 0, 0)
        target_band.FlushCache()
        target_ds = None
        print(RES, c, v)


def check(ls):
    for ll in ls:
        ds0 = ll[0]
        ds1 = ll[1]
        assert ds0.RasterXSize == ds1.RasterXSize, "Raster X size should be equal"
        assert ds0.RasterYSize == ds1.RasterYSize, "Raster X size should be equal"
        assert ds0.GetGeoTransform() == ds1.GetGeoTransform(), "Geotransform should be equal"
        # assert ds0.GetProjection() == ds1.GetProjection(), "Projection should be equal"


def stats(band):
    stats = band.GetStatistics(True, True)
    if stats is not None:
        print("[ STATS ] =  Minimum=%.3f, Maximum=%.3f, Mean=%.3f, StdDev=%.3f" % (
            stats[0], stats[1], stats[2], stats[3]))


def removecomplexpart(data):
    if data.dtype == np.dtype('complex128'):
        data = data.astype('float128')
    if data.dtype == np.dtype('complex64'):
        data = data.astype('float64')
    return data


if __name__ == '__main__':
    map = "map prefix"
    model = "model prefix"
    sourcefolder = 'Source Folder'
    densityfolder = 'Density Folder'
    createDensity(map, sourcefolder, densityfolder, model)
