# Detection of Potential Geothermal Exploration Sites from Hyperspectral Images via Deep Learning
# Author: Ebubekir Demir <edemir@mines.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.


from __future__ import print_function
import time
from fastshap import Surrogate, FastSHAP
start = time.time()
from datetime import timedelta
import numpy as np
import doe_tiff as dt
import argparse
from tqdm import tqdm

import matplotlib
from numpy import argmax
import os

from sklearn.preprocessing import LabelBinarizer
from sklearn.model_selection import train_test_split

# import the necessary packages

import tensorflow as tf
from keras.layers import Input
from keras.layers import Dense
from keras.models import Model

from tensorflow.keras.utils import to_categorical

#from fastshap import FastSHAP

# Default values
SAMPLES_TO_CREATE = 1200
KERNEL_PIXELS = 17
KERNEL_CHANNELS = 5
DEFAULT_DIRECTORY = 'dataset'

matplotlib.use("Agg")
DEFAULT_WEIGHTS_FILE_NAME = 'doe_cnn.weights'
DEFAULT_MODEL_FILE_NAME = 'doe_cnn.model'
DEFAULT_LABEL_FILE_NAME = 'doe_cnn.labels.pickle'

INIT_LR = 5e-1  # Default Loss Rate
INIT_DECAY = 1e-3  # Default Decay
CHANNELS = 5  # This will be redefined based on parameters
KERNEL_PIXELS = 17  # Default pixels by side on each tile
NUM_CLASSES = 2  # Default number of classes ("Geothermal", "Non-Geothermal")
BS = 32  # Default batch size
EPOCHS = 500

try:
    raw_input  # Python 2
except NameError:
    raw_input = input  # Python 3

print('Set-up complete.')

if __name__ == '__main__':
    ''' Main instructions '''
    print('Parsing input...')
    ap = argparse.ArgumentParser()

    ap.add_argument("-i", "--image", required=True, help="path to input multi-band image (i.e., image file name)")
    ap.add_argument("-i2", "--image2", required=False,
                    help="path to second input multi-band image (i.e., image file name) (optional)", default=None)
    ap.add_argument("-i3", "--image3", required=False,
                    help="path to third input multi-band image (i.e., image file name) (optional)", default=None)
    ap.add_argument("-s", "--samples", required=True, help="number of samples to create", type=int)
    ap.add_argument("-b", "--batch_size", required=True, help="Defines batch size", type=int)
    ap.add_argument("-e", "--epochs", required=True, type=int, help="Number of epochs to train")
    ap.add_argument("-c", "--channels", required=False, help='Number of channels in each image', default=CHANNELS,
                    type=int)
    ap.add_argument("-k", "--kernel_size", required=False, help='Number of pixels by side in each image',
                    default=KERNEL_PIXELS, type=int)
    ap.add_argument("-m", "--model", required=True, help="path to input model (h5)")
    ap.add_argument("-d", "--directory", required=True, help="directory")

    args = vars(ap.parse_args())
    image_name = args["image"]
    image_name2 = args["image2"]
    image_name3 = args["image3"]

    num_samples = args["samples"]
    batch_size = args["batch_size"]
    num_epochs = args["epochs"]

    num_channels = args["channels"]
    image_channels = num_channels - 1

    kernel_size = args["kernel_size"]
    kernel_pixels = kernel_size
    model_file = args["model"]
    model_dir = args["directory"]

    augment_data = None
    reset_model = "reset"
    num_classes = NUM_CLASSES
    true_random = None
    validate_only = None
    weights_file = None
    output_curves_file = None

    PADDING = int(kernel_size / 2)
    img_b = dt.io.read_gdal_file(image_name)
    # img_b = dt.io.read_gdal_file(image_name)
    max_channels = img_b.shape[2]
    assert num_channels > 0, 'Channels has to be a positive integer'
    assert num_channels <= max_channels, 'Channels (' + str(num_channels) + ') has to be equal or lower than ' + str(
        max_channels)
    img_b_scaled = img_b[:, :, 1:num_channels + 1]
    print(img_b_scaled.shape)
    mask_b = img_b[:, :, 0]  # first band is Ground Truth
    for i in range(0, img_b_scaled.shape[2]):
        print("band (", i, ") min:", img_b_scaled[:, :, i].min())
        print("band (", i, ") max:", img_b_scaled[:, :, i].max())
    img_b_scaled = dt.frame_image(img_b_scaled, PADDING)
    sc = dt.GeoTiffConvolution(img_b_scaled, kernel_size, kernel_size)

    new_map = np.zeros_like(mask_b)  # creates empty map
    data = []
    labels = []
    # Land classes go from 0 to 1
    for land_type_class in range(0, 2):
        mask_b_class = np.array(np.where(mask_b == land_type_class)).T
        print('Items in class ', land_type_class, ': ', len(mask_b_class))
        mask_b_scaled = mask_b_class + [PADDING, PADDING]
        ln = len(mask_b_scaled)
        choices = np.random.choice(len(mask_b_scaled), num_samples, replace=False)
        for c in tqdm(choices, desc="Calculating...", ascii=False, ncols=75):
            x = mask_b_scaled[c][0] - PADDING
            y = mask_b_scaled[c][1] - PADDING
            new_map[x, y] = 1 + land_type_class
            my_slice = sc.apply_mask(mask_b_scaled[c][0], mask_b_scaled[c][1])
            data.append(my_slice)
            labels.append(land_type_class)

    if image_name2 is not None:
        img_b = dt.io.read_gdal_file(image_name2)
        #img_b = np.load(image_name2)
        max_channels = img_b.shape[2]
        assert num_channels > 0, 'Channels has to be a positive integer'
        assert num_channels <= max_channels, 'Channels has to be equal or lower than {}'.format(max_channels)
        img_b_scaled = img_b[:, :, 1:num_channels + 1]
        mask_b = img_b[:, :, 0]  # first band is Ground Truth
        for i in range(0, img_b_scaled.shape[2]):
            print("band (", i, ") min:", img_b_scaled[:, :, i].min())
            print("band (", i, ") max:", img_b_scaled[:, :, i].max())
        img_b_scaled = dt.frame_image(img_b_scaled, PADDING)
        sc = dt.GeoTiffConvolution(img_b_scaled, kernel_size, kernel_size)

        new_map = np.zeros_like(mask_b)  # creates empty map
        # Land classes go from 0 to 1
        for land_type_class in range(0, 2):
            mask_b_class = np.array(np.where(mask_b == land_type_class)).T
            print('Items in class ', land_type_class, ': ', len(mask_b_class))
            mask_b_scaled = mask_b_class + [PADDING, PADDING]
            ln = len(mask_b_scaled)
            choices = np.random.choice(len(mask_b_scaled), num_samples, replace=False)
            for c in tqdm(choices, desc="Calculating...", ascii=False, ncols=75):
                x = mask_b_scaled[c][0] - PADDING
                y = mask_b_scaled[c][1] - PADDING
                new_map[x, y] = 1 + land_type_class
                my_slice = sc.apply_mask(mask_b_scaled[c][0], mask_b_scaled[c][1])
                data.append(my_slice)
                labels.append(land_type_class)

    if image_name3 is not None:
        img_b = dt.io.read_gdal_file(image_name3)
        #img_b = np.load(image_name3)
        max_channels = img_b.shape[2]
        assert num_channels > 0, 'Channels has to be a positive integer'
        assert num_channels <= max_channels, 'Channels has to be equal or lower than {}'.format(max_channels)
        img_b_scaled = img_b[:, :, 1:num_channels + 1]
        mask_b = img_b[:, :, 0]  # first band is Ground Truth
        for i in range(0, img_b_scaled.shape[2]):
            print("band (", i, ") min:", img_b_scaled[:, :, i].min())
            print("band (", i, ") max:", img_b_scaled[:, :, i].max())
        img_b_scaled = dt.frame_image(img_b_scaled, PADDING)
        sc = dt.GeoTiffConvolution(img_b_scaled, kernel_size, kernel_size)

        new_map = np.zeros_like(mask_b)  # creates empty map
        # Land classes go from 0 to 1
        for land_type_class in range(0, 2):
            mask_b_class = np.array(np.where(mask_b == land_type_class)).T
            print('Items in class ', land_type_class, ': ', len(mask_b_class))
            mask_b_scaled = mask_b_class + [PADDING, PADDING]
            ln = len(mask_b_scaled)
            choices = np.random.choice(len(mask_b_scaled), num_samples, replace=False)
            for c in tqdm(choices, desc="Calculating...", ascii=False, ncols=75):
                x = mask_b_scaled[c][0] - PADDING
                y = mask_b_scaled[c][1] - PADDING
                new_map[x, y] = 1 + land_type_class
                my_slice = sc.apply_mask(mask_b_scaled[c][0], mask_b_scaled[c][1])
                data.append(my_slice)
                labels.append(land_type_class)

    IMAGE_DIMS = (kernel_pixels, kernel_pixels, image_channels)
    BATCH_DIMS = (None, kernel_pixels, kernel_pixels, image_channels)

    print('Read images:', len(data))
    # scale the raw pixel intensities to the range [0, 1]
    data = np.asarray(data, dtype=np.float32)
    data = np.nan_to_num(data)
    labels = np.array(labels)
    print("[INFO] data matrix: {:.2f}MB".format(
        data.nbytes / (1024 * 1024.0)))
    # binarize the labels
    lb = LabelBinarizer()
    labels_lb = lb.fit_transform(labels)

    y_binary = to_categorical(labels, num_classes=num_classes)
    y_inverse = argmax(y_binary, axis=1)
    print('Read labels:', (lb.classes_))
    print('Transformed labels:', (y_binary[:10]))
    print('Inverted labels:', (y_inverse[:10]))
    print('Any nulls?: ', np.isnan(data).any())
    ###
    '''
    Creates the network
    '''
    # Define distribution strategy
    strategy = tf.distribute.MirroredStrategy()

    # construct model under distribution strategy scope
    with strategy.scope():

        model = tf.keras.models.load_model(os.path.join(model_dir, model_file))
        # partition the data into training and testing splits using 20% of
        # the data for training and the remaining 80% for testing/validation
        (trainX, testX, trainY, testY) = train_test_split(data,
                                                          y_binary, test_size=0.8, random_state=42)
        (validateX, testX, validateY, testY) = train_test_split(testX, testY,
                                                                test_size=0.75, random_state=42)

        num_features = kernel_pixels * kernel_pixels * image_channels
        model_input = Input(shape=IMAGE_DIMS, name='input')
        net = Dense(128, activation="relu")(model_input)
        net = Dense(128, activation="relu")(net)
        out = Dense(num_classes, activation="softmax")(net)
        surrogate_model = Model(model_input, out)
        surrogate = Surrogate(surrogate_model=surrogate_model, num_features=IMAGE_DIMS, baseline=0)

        surrogate.train(original_model=model,
                        train_data=trainX,
                        val_data=validateX,
                        batch_size=64,
                        max_epochs=100,
                        validation_batch_size=10000,
                        loss_fn="categorical_crossentropy",
                        lr=1e-3,
                        min_lr=1e-5,
                        lr_factor=0.9,
                        lookback=10,
                        verbose=1,
                        model_dir=model_dir)

        model_input = Input(shape=num_features, name='input')
        net = Dense(128, activation="relu")(model_input)
        net = Dense(128, activation="relu")(net)
        out = Dense(num_classes * num_features, activation="linear")(net)

        explainer = Model(model_input, out)

        fastshap = FastSHAP(explainer, surrogate, normalization='additive', link='identity')

        fastshap.train(train_data=trainX,
                       val_data=validateX,
                       batch_size=32,
                       validation_batch_size=64,
                       num_samples=8,
                       max_epochs=200,
                       model_dir=model_dir,
                       lr=2e-4,
                       verbose=1,
                       lookback=5)

        fastshap_values = fastshap.shap_values(x)[0].transpose(1, 0)
        np.save("shaptest_values.npy")
        print(fastshap_values.shape)

end = time.time()
td = timedelta(seconds=int(end - start))
print("Total processing time", td)
