# Detection of Potential Geothermal Exploration Sites from Hyperspectral Images via Deep Learning
# Author: Ebubekir Demir <edemir@mines.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

import argparse

import numpy as np
from osgeo import gdal
from osgeo.gdalconst import *
from osgeo import gdal_array


def extractData(rasterfile, npydestination):
    ds = gdal.Open(rasterfile)
    xsize = ds.RasterXSize
    ysize = ds.RasterYSize
    print("Raster Size:", xsize, ysize)
    count = ds.RasterCount
    print("Band Count:", count)
    gettr = ds.GetGeoTransform()
    print("Raster Geotransform:", gettr)
    crs = ds.GetProjection()
    print("Raster CRS:", crs)
    my_image = gdal_array.LoadFile(rasterfile)
    my_image = np.ma.masked_invalid(my_image)
    if count == 1:
        my_image = my_image.T
    else:
        my_image = np.transpose(my_image, axes=(2, 1, 0))
    np.save(npydestination, my_image.data)


def createRaster(inimage, outimage, datadestination, datatype):
    data = np.load(datadestination)
    inDs = gdal.Open(inimage)
    ysize = inDs.RasterYSize
    xsize = inDs.RasterXSize
    print(data.shape, ysize, xsize)
    gettr = inDs.GetGeoTransform()
    print("Raster Geotransform:", gettr)
    crs = inDs.GetProjection()
    print("Raster CRS:", crs)
    assert (ysize, xsize) == data.shape, "data and raster extend do not match"
    driver = inDs.GetDriver()
    outDs = None
    if datatype == "int":
        outDs = driver.Create(outimage, xsize, ysize, 1, GDT_Int32)
    elif datatype == "float":
        outDs = driver.Create(outimage, xsize, ysize, 1, GDT_Float32)
    else:
        print("Wrong Data Type. Supported Data Types: int and float")
        return
    outBand = outDs.GetRasterBand(1)
    outData = data
    outBand.WriteArray(outData, 0, 0)
    outBand.FlushCache()
    outDs.SetGeoTransform(inDs.GetGeoTransform())
    outDs.SetProjection(inDs.GetProjection())
    outDs.FlushCache()
    del outData
    outDs = None
    print("saving file: ", outimage)


if __name__ == '__main__':
    ''' Main instructions '''
    print('Parsing input...')
    ap = argparse.ArgumentParser()
    ap.add_argument("-e", "--extract", required=False, help="flag to extract data", action='store_true', default=False)
    ap.add_argument("-c", "--create", required=False, help="flag to create raster", action='store_true', default=False)
    ap.add_argument("-i", "--inputraster", required=True, help='input raster image', default=None, type=str)
    ap.add_argument("-d", "--data", required=True, help='numpy data file', default=None, type=str)
    ap.add_argument("-t", "--type", required=False, type=str, help="output data type (int or float)", default=None)
    ap.add_argument("-o", "--outputraster", required=False, type=str, help="output rasterfile", default=None)
    args = vars(ap.parse_args())
    extract = args["extract"]
    create = args["create"]
    inputraster = args["inputraster"]
    data = args["data"]
    type = args["type"]
    outputraster = args["outputraster"]

    if extract:
        extractData(inputraster, data)
    elif create:
        createRaster(inputraster, outputraster, data, type)
