# Detection of Potential Geothermal Exploration Sites from Hyperspectral Images via Deep Learning
# Author: Ebubekir Demir <edemir@mines.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from osgeo import gdal, ogr
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.metrics import roc_curve, auc
import os


def linedistance(densitymapfolder, shapefile, densityvalcsvfolder, fprtprthresholdscsvfolder, rocfolder):
    for filename in os.listdir(densitymapfolder):
        file = os.path.join(densitymapfolder, filename)
        ds = gdal.Open(file)
        xsize = ds.RasterXSize
        ysize = ds.RasterYSize
        gettr = ds.GetGeoTransform()
        pixel_size = gettr[1]
        xref = gettr[0]
        yref = gettr[3]
        crs = ds.GetProjection()
        band = ds.GetRasterBand(1)
        dendata = band.ReadAsArray()
        if dendata.dtype == np.dtype('complex128'):
            dendata = dendata.astype('float128')
        if dendata.dtype == np.dtype('complex64'):
            dendata = dendata.astype('float64')
        driver = ogr.GetDriverByName("ESRI Shapefile")
        source = driver.Open(shapefile, 1)
        layer = source.GetLayer()
        ldefn = layer.GetLayerDefn()
        feature_id = 0
        print(file)
        cols = ["M_Alunite", "M_Chalcedo", "M_Opal", "M_Kaolinit", "M_Hematite", "M_Epidote"]
        for colname in cols:
            print(colname)
            for n in range(ldefn.GetFieldCount()):
                fdefn = ldefn.GetFieldDefn(n)
                if fdefn.name == colname:
                    feature_id = n
                    break
            truth = []
            density = []
            name = []
            for feature in layer:
                geometry = feature.GetGeometryRef()
                if geometry.GetGeometryType() == ogr.wkbPoint:
                    x, y, z = geometry.GetPoint()
                    x, y = getpixelcoord(x, y, xref, yref, pixel_size)
                    if x != 0 and y != 0:
                        res = dendata[y][x]
                        truth.append(feature.GetField(feature_id))
                        name.append(feature.GetField(0))
                        density.append(res)

            dict_val = {"Name": name, "Density": density, colname: truth}
            df_val = pd.DataFrame(dict_val)
            denfile = os.path.join(densityvalcsvfolder, filename) + "_" + colname + ".csv"
            df_val.to_csv(denfile)
            print("Density data exported:", denfile)
            fpr, tpr, thresholds = roc_curve(truth, density)
            roc_auc = auc(fpr, tpr)
            plt.figure()
            lw = 2
            plt.plot(
                fpr,
                tpr,
                color="red",
                lw=lw,
                label="Model Fusion (auc=%0.3f)" % roc_auc,
            )
            plt.plot([0, 1], [0, 1], color="navy", lw=lw, linestyle="--")
            plt.xlim([0.0, 1.0])
            plt.ylim([0.0, 1.05])
            plt.xlabel("False Positive Rate")
            plt.ylabel("True Positive Rate")
            plt.title("Coso Map ROC")
            plt.legend(loc="lower right")
            denfile = os.path.join(rocfolder, filename) + "_" + colname + ".png"
            plt.savefig(denfile)
            plt.show()
            plt.close('all')

            dict_val = {"FPR": fpr, "TPR": tpr, "Thresholds": thresholds}
            df_val = pd.DataFrame(dict_val)
            roccsvfile = os.path.join(fprtprthresholdscsvfolder, filename) + "_" + colname + ".csv"
            df_val.to_csv(roccsvfile)
            print("ROC data - FPR vs TPR exported:", roccsvfile)


def getpixelcoord(x, y, xref, yref, size):
    if x < xref:
        print("error")
        return 0, 0
    if y > yref:
        print("error")
        return 0, 0
    xval = (int)((x - xref) / size)
    yval = (int)((yref - y) / size)
    return xval, yval


if __name__ == '__main__':
    linedistance(densityfolder, shp, dencsv, roccsv, rocpng)
