# Detection of Potential Geothermal Exploration Sites from Hyperspectral Images via Deep Learning
# Author: Ebubekir Demir <edemir@mines.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.


from __future__ import print_function
import time
start = time.time()
from datetime import timedelta
import argparse
import tensorflow as tf
from keras.models import load_model
from tqdm import tqdm
import matplotlib
import numpy as np
import os.path as path
import doe_tiff2 as dt

# Setup matplotlib
matplotlib.use("Agg")
# Create defaults for file names
DEFAULT_WEIGHTS_FILE_NAME = 'doe_cnn.weights'
DEFAULT_MODEL_FILE_NAME = 'doe_cnn.model'
DEFAULT_LABEL_FILE_NAME = 'doe_cnn.labels.pickle'
# Define defaults for input file parameters
CHANNELS = 5  # This will be redefined based on parameters
KERNEL_PIXELS = 17  # Default pixels by side on each tile
NUM_CLASSES = 2  # Default number of classes ("Geothermal", "Non-Geothermal")
# Default hyperparameters
INIT_LR = 5e-1  # Default Loss Rate
INIT_DECAY = 1e-3  # Default Decay
BS = 32
EPOCHS = 500
# Default labels
class2code = {'none': 0,
              'Non-geothemal': 1,
              'Geothemal': 2}
code2class = {0: 'none',
              1: 'Non-geothemal',
              2: 'Geothemal'}

# Ensure Python 3 is being called
try:
    raw_input  # Python 2
    print("Run in Python 3")
    raise SystemExit
except NameError:
    raw_input = input  # Python 3



def predict(model_file, image_name, num_channels, kernel_size, prediction_map, probability_map):
    PADDING = int(kernel_size / 2)
    # Ensures model file exists and is really a file
    try:
        assert path.exists(model_file), 'Model path {} does not exist'.format(model_file)
        assert path.isfile(model_file), 'Model file {} is not a file'.format(model_file)
    except:
        raise FileNotFoundError
    try:
        assert path.isfile(image_name), 'Image file {}: is not a file'.format(model_file)
        #img_b = dt.io.read_gdal_file(image_name)
        img_b = np.load(image_name)
        max_channels = img_b.shape[2]
    except:
        print("Image file not found or erroneous")
        raise FileNotFoundError
    # Read model file
    print('[INFO] Loading model from file...')

    # Define distribution strategy
    strategy = tf.distribute.MirroredStrategy()
    # construct model under distribution strategy scope
    with strategy.scope():
        model3 = load_model(model_file)
        model3.summary()
        # Get rid of NaN's
        img_b = np.array(img_b, dtype=np.float64)
        img_b = np.nan_to_num(img_b)
        print("Image shape:", img_b.shape)
        assert num_channels > 0, 'Channels has to be a positive integer'
        assert num_channels <= max_channels, 'Channels has to be equal or lower than {}'.format(max_channels)
        img_b_scaled = img_b[:, :, 1:num_channels+1]
        print("DLM input image shape:", img_b_scaled.shape)
        mask_b = img_b[:, :, 0]  # first band is Ground Truth
        (img_x, img_y) = mask_b.shape
        print("Mask shape:", mask_b.shape)
        new_map = np.empty((img_x, img_y, 2))
        print("Map shape:", new_map.shape)
        for i in range(0, img_b_scaled.shape[2]):
            print("band (", i, ") min:", img_b_scaled[:, :, i].min())
            print("band (", i, ") max:", img_b_scaled[:, :, i].max())
        img_b_scaled = dt.frame_image(img_b_scaled, PADDING)
        sc = dt.GeoTiffConvolution(img_b_scaled, kernel_size, kernel_size)

        print('[INFO] Creating prediction map...')
        for i in tqdm(range(img_x), desc="Predicting...", ascii=False, ncols=75):
            data = []
            for j in range(img_y):
                image = sc.apply_mask(i + PADDING, j + PADDING)
                data.append(image)
            data = np.array(data, dtype=np.float64)
            data = np.nan_to_num(data)
            pre_y = model3.predict(data, verbose=0, batch_size=10000)
            # pre_y = pre_y.argmax(axis=-1)
            new_map[i, :, :] = pre_y
        new_map = np.asarray(new_map)
        #saveraster(image_name, probability_map, new_map[:, :, 1].T, GDT_CFloat32)
        np.save(probability_map, new_map[:, :, 1].T)
        preddata = new_map.argmax(axis=-1).T
        #saveraster(image_name, prediction_map, preddata, GDT_Int32)
        np.save(prediction_map, preddata)

print('Set-up complete.')

''' Main program '''
if __name__ == '__main__':
    ''' Main instructions '''
    print('Parsing input...')
    ap = argparse.ArgumentParser()
    ap.add_argument("-c", "--channels", required=False, help='Number of channels in each image', default=CHANNELS,
                    type=int)
    ap.add_argument("-i", "--image", required=True, help="path to input multi-band image (i.e., image file name)")
    ap.add_argument("-k", "--kernel_size", required=False, help='Number of pixels by side in each image',
                    default=KERNEL_PIXELS, type=int)
    ap.add_argument("-m", "--model", required=True, help="path to output model")
    ap.add_argument("-n", "--num_classes", required=False, help='Number of classes', default=NUM_CLASSES, type=int)
    ap.add_argument("-p", "--prediction_map", required=True, type=str, help="Prediction Raster file")
    ap.add_argument("-o", "--probability_map", required=True, type=str, help="Probability Raster file")

    args = vars(ap.parse_args())
    num_channels = args["channels"]
    image_name = args["image"]
    kernel_size = args["kernel_size"]
    model_file = args["model"]
    num_classes = args["num_classes"]
    prediction_map = args["prediction_map"]
    probability_map = args["probability_map"]
    predict(model_file, image_name, num_channels, kernel_size, prediction_map, probability_map)

end = time.time()
td = timedelta(seconds=int(end - start))
print("Total processing time", td)
